==> tecfy_corev2_package-1.1.44.tar.gz: lib/controllers/entity/entity_form.controller.dart <==
part of tecfy_corev2_client_package;

class EntityFormController extends BaseController {
  /// entity object
  late Entity entity;

  /// the display field anme
  late String titleField;

  /// second display field name
  late String subTitleField;

  /// tail field name
  late String tailField;

  /// data
  Map<String, dynamic>? data;
  Map<String, dynamic>? dataOld;

  /// properties
  late List<Property> properties;
  late List<ViewTab> tabs;

  /// properties need calculation
  Map<String, Property> calcProps = {};

  /// is new object
  bool isAddNew = false;

  /// is true, the controller will load data from server on open
  bool loadData = false;

  /// form tabs
  //late Map<String, List<List<Property>>> tabs;

  /// is can save (Add/Update)
  bool get canSave =>
      !_readOnly &&
      (data != null && data!['isLocked'] != true) &&
      ((isAddNew && entity.permission.canAdd) ||
          (!isAddNew && entity.permission.canUpdate)) &&
      (entity.entityType != EntityTypes.SystemGanarated &&
          entity.entityType != EntityTypes.ViewOnly &&
          entity.entityType != EntityTypes.Lookup);
  bool _readOnly;

  /// is read only
  bool get readOnly => !canSave;
  List<dynamic>? _entitySaveConfirm;

  EntityFormController(State<StatefulWidget> state, String entityName,
      this.data, this._readOnly, this.loadData)
      : super(state) {
    busy = false;
    isAddNew = data == null || data!['_id'] == null;
    if (kDebugMode) print('-------------- EntityFormController => $entityName');
    entity = EntityService.getEntity(entityName) ?? Entity.fromMap({});
    if (kDebugMode)
      print(
          'Entity ${entity.dbName} Can Add: ${entity.permission.canAdd} Can Edit: ${entity.permission.canUpdate}');
    properties =
        entity.properties.where((p) => isAddNew ? p.isAdd : p.isEdit).toList();
    //properties.forEach((p) => p.isVisable = true);

    for (var prop in properties.where((p) =>
        p.propertyType == PropertyTypes.Number &&
        (p as PropertyNumber).calcFormula != null &&
        p.calcFormula!.isNotEmpty)) {
      calcProps[prop.dbName] = prop;
    }

    if (data == null) {
      data = Map<String, dynamic>();
    }

    print('////////////////////////////');
    print('entity.tabs.length ${entity.tabs.length}');
    print(properties.map((e) => e.editTabName));

    tabs = entity.tabs
        .where((t) => properties.any((e) => e.editTabName == t.dbName))
        .toList();

    loadFromDb();
  }

  Future loadFromDb() async {
    if (entity.isLocal || !loadData) return;
    entity.updatedProps.clear();
    if (data == null || data!['_id'] == null) {
      await onValueChanged('', '', true); // for default values
      return;
    }
    await Future.delayed(Duration(
        milliseconds:
            30)); // delay a bit because you cannot show "loading" before page initlization
    setLoading();
    var result = await EntityService.getOne(entity.dbName, data!['_id'],
        forView: readOnly, withRelatedData: true);
    setResume();
    if (result != null) {
      data = result;
      dataOld = json
          .decode(json.encode(result, toEncodable: EntityService.customEncode));
      if (kDebugMode) print(data);
      await onValueChanged('', '', true);
      setState();
    }
  }

  List<Property> getProps(ViewTab tab) {
    return properties.where((e) => e.editTabName == tab.dbName).toList();
  }

  Future onValueChanging(value, propName) async {
    MapService.updateValue(data!, propName, value);
    if (kDebugMode) print('-----------* value changed ! ($propName)');
    entity.updatedProps.clear();
    var property = properties.firstWhere((p) => p.dbName == propName);
    if (property.propertyType == PropertyTypes.Dropdown) {
      var prop = property as PropertyDropdown;
      if (MapService.getValue(data!, property.dbName) == null) {
        setState();
        return;
      }
      var props = prop.overwritePropertiesOnSelect?.keys ?? [];
      for (var key in props) {
        await onValueChanged(MapService.getValue(data!, prop.dbName)[key],
            prop.overwritePropertiesOnSelect![key]);
      }
    }

    if (await onValueChanged(value, propName)) {
      setState();
    }
  }

  Future<bool> onValueChanged(value, propName, [bool calcAll = false]) async {
    MapService.updateValue(data!, propName, value);
    if (readOnly) return false;
    if (kDebugMode) print('-----------Calc ($propName) $calcAll');
    try {
      if (entity.updatedProps[propName] != null) return false;
      entity.updatedProps[propName] = true;
      if (!calcAll) MapService.updateValue(data!, propName, value);

      bool updated = await entity.setDefaultValues(data!);

      // for (Property p in properties) {
      //   var r = await p.isVisible(data, entity);
      //   if (r != p.isVisable) {
      //     p.isVisable = r;
      updated = true;
      //   }
      // }
      // check the calced properties
      for (PropertyNumber p in calcProps.values
          .where((e) => e.propertyType == PropertyTypes.Number)
          .toList()
          .cast<PropertyNumber>()) {
        if (p.formulaType == "Formula") {
          if (calcAll ||
              p.calcedProps.contains(p.upperFirst(propName)) ||
              p.calcedProps.contains(propName)) {
            var vars = p.calcedProps
                .map((pro) => data![p.lowerFirst(pro)] ?? 0) //TODO: fix dots
                .toList();
            await onValueChanged(p.getCalcResult(vars), p.dbName);
            //setState();
            if (kDebugMode) print('calc ${p.dbName}');
          }
        } else if (p.formulaType == "SumSub") {
          if (calcAll || p.sumEntityName == propName) {
            double val = 0;
            if (data![p.sumEntityName] != null && //TODO: fix dots
                data![p.sumEntityName] is List) {
              //TODO: fix dots
              for (var row in data![p.sumEntityName]) {
                //TODO: fix dots
                val += row[p.sumPropertyName] ?? 0;
              }
            }
            await onValueChanged(val, p.dbName);
            //setState();
          }
        }
      }

      // if (updated // || (!calcAll && properties.any((p) => p.filterValueField == propName))
      //     ) setState();
      return updated;
    } catch (err, t) {
      print('XXXXXX ERROR onValueChanged: $err : $t');
    }
    return false;
  }

  Future<bool> onSave() async {
    if (readOnly) return false;
    // show loading
    setLoading();
    // if (!isAddNew && dataOld != null) {
    //   // remove unchanged
    //   for (var k in dataOld!.keys) {
    //     if (k != '_id' && dataOld![k] == data![k]) {
    //       //TODO: fix dots
    //       data!.remove(k);
    //     }
    //   }
    // }
    print(data);
    var result = await EntityService.saveEntity(entity, data, context, isAddNew,
        entitySaveConfirm: _entitySaveConfirm);
    setResume();
    if (!result.success) {
      try {
        if (kDebugMode) print(result.data);
        var question = '';
        if (result.data != null && result.data.length > 0) {
          question = result.data[0]['nameAr'] ?? '';
        }
        if (question.isNotEmpty) {
          var r = await AppNavigator.confirmMessage(
              context, result.message ?? '',
              message: question);
          if (!r) return false;
          result.data[0]['value'] = true;
          _entitySaveConfirm = result.data;
          return await onSave();
        } else {
          AppNavigator.showMessage(context,
              result.message ?? '${result.innerError}', MessageType.error,
              message: question);
          print(result);
        }
      } catch (err, t) {
        await AppNavigator.showMessage(context,
            result.message ?? '${result.innerError}', MessageType.error,
            message: result.message);
        print(err);
        print(t);
      }
      return false;
    }
    ConfigService.setValueMap("savedDataForm${entity.name}", {});
    // show message
    await AppNavigator.showMessage(
        context, 'تم الحفظ بنجاح', MessageType.success);
    AppNavigator.pop(context, true);
    return true;
  }

  Future<bool> onSaveNew() async {
    if (readOnly) return false;
    // show loading
    isAddNew = true;
    resetIds(data!);
    return await onSave();
  }

  Future onOpenHistory() {
    if (data == null) return Future.delayed(Duration.zero);
    return AppNavigator.navigateTo(context,
        () => EntityHistoryPage(entity.dbName, data?['_id'] ?? '', data: data));
  }

  resetIds(Map<String, dynamic> dat) {
    dat['_id'] = null;
    dat['recordNumber'] = null;
    for (var k in dat.keys) {
      if (dat[k] is List) {
        if (kDebugMode) print('d---------$k');
        for (var d in dat[k]) {
          resetIds(d);
        }
      }
    }
  }

  lowerFirst(String s) {
    return '${s[0].toLowerCase()}${s.substring(1)}';
  }

//   void initTabs() {
//     this.tabs = {};
//     for (var prop in this.entity.properties) {
//       if (isAddNew ? prop.isAdd : prop.isEdit) {
//         if (prop.editTabName == null || (prop.editTabName?.isEmpty ?? true))
//           prop.editTabName = AppLocale.translate('Details');

// //prop.propertyType == PropertyTypes.OtherEntityMany
//         if (prop.size == PropertySizes.NewTab) {
//           this.tabs[prop.name] = [
//             [prop]
//           ];
//         } else {
//           if (this.tabs.length == 0)
//             this.tabs[prop.editTabName ?? ''] = [
//               [prop]
//             ];
//           else {
//             if (this.tabs[prop.editTabName] == null) {
//               this.tabs[prop.editTabName ?? ''] = [
//                 [prop]
//               ];
//             } else {
//               if (prop.fullWidth)
//                 this.tabs[prop.editTabName]?.add([prop]);
//               else
//                 this.tabs[prop.editTabName]?.last.add(prop);
//             }
//           }
//         }
//       }
//       if (prop.newLineAfter) this.tabs[prop.editTabName]?.add([]);
//     }

//     for (var tk in tabs.keys) {
//       tabs[tk] = tabs[tk]?.where((e) => e.length > 0).toList() ?? [];
//       if (tabs[tk]?.length == 0) {
//         tabs.remove(tk);
//       }
//     }
//   }
}

==> tecfy_corev2_package-1.1.51.tar.gz: lib/controllers/entity/entity_form.controller.dart <==
part of tecfy_corev2_client_package;

class EntityFormController extends BaseController {
  /// entity object
  late Entity entity;

  /// the display field anme
  late String titleField;

  /// second display field name
  late String subTitleField;

  /// tail field name
  late String tailField;

  /// data
  Map<String, dynamic>? data;
  Map<String, dynamic>? dataOld;

  /// properties
  late List<Property> properties;
  late List<ViewTab> tabs;

  /// properties need calculation
  Map<String, Property> calcProps = {};

  /// is new object
  bool isAddNew = false;

  /// is true, the controller will load data from server on open
  bool loadData = false;

  /// form tabs
  //late Map<String, List<List<Property>>> tabs;

  /// is can save (Add/Update)
  bool get canSave =>
      !_readOnly &&
      (data != null && data!['isLocked'] != true) &&
      ((isAddNew && entity.permission.canAdd) ||
          (!isAddNew && entity.permission.canUpdate)) &&
      (entity.entityType != EntityTypes.SystemGanarated &&
          entity.entityType != EntityTypes.ViewOnly &&
          entity.entityType != EntityTypes.Lookup);
  bool _readOnly;

  /// is read only
  bool get readOnly => !canSave;
  List<dynamic>? _entitySaveConfirm;

  EntityFormController(State<StatefulWidget> state, String entityName,
      this.data, this._readOnly, this.loadData)
      : super(state) {
    busy = false;
    isAddNew = data == null || data!['_id'] == null;
    if (kDebugMode) print('-------------- EntityFormController => $entityName');
    entity = EntityService.getEntity(entityName) ?? Entity.fromMap({});
    if (kDebugMode)
      print(
          'Entity ${entity.dbName} Can Add: ${entity.permission.canAdd} Can Edit: ${entity.permission.canUpdate}');
    properties =
        entity.properties.where((p) => isAddNew ? p.isAdd : p.isEdit).toList();
    //properties.forEach((p) => p.isVisable = true);

    for (var prop in properties.where((p) =>
        p.propertyType == PropertyTypes.Number &&
        (p as PropertyNumber).calcFormula != null &&
        p.calcFormula!.isNotEmpty)) {
      calcProps[prop.dbName] = prop;
    }

    if (data == null) {
      data = Map<String, dynamic>();
    }

    print('////////////////////////////');
    print('entity.tabs.length ${entity.tabs.length}');
    print(properties.map((e) => e.editTabName));

    tabs = entity.tabs
        .where((t) => properties.any((e) => e.editTabName == t.dbName))
        .toList();

    loadFromDb();
  }

  Future loadFromDb() async {
    if (entity.isLocal || !loadData) return;
    entity.updatedProps.clear();
    if (data == null || data!['_id'] == null) {
      await onValueChanged('', '', true); // for default values
      return;
    }
    await Future.delayed(Duration(
        milliseconds:
            30)); // delay a bit because you cannot show "loading" before page initlization
    setLoading();
    var result = await EntityService.getOne(entity.dbName, data!['_id'],
        forView: readOnly, withRelatedData: true);
    setResume();
    if (result != null) {
      data = result;
      dataOld = json
          .decode(json.encode(result, toEncodable: EntityService.customEncode));
      if (kDebugMode) print(data);
      await onValueChanged('', '', true);
      setState();
    }
  }

  List<Property> getProps(ViewTab tab) {
    return properties.where((e) => e.editTabName == tab.dbName).toList();
  }

  Future onValueChanging(value, propName) async {
    MapService.updateValue(data!, propName, value);
    if (kDebugMode) print('-----------* value changed ! ($propName)');
    entity.updatedProps.clear();
    var property = properties.firstWhere((p) => p.dbName == propName);
    if (property.propertyType == PropertyTypes.Dropdown) {
      var prop = property as PropertyDropdown;
      if (MapService.getValue(data!, property.dbName) == null) {
        setState();
        return;
      }
      var props = prop.overwritePropertiesOnSelect?.keys ?? [];
      for (var key in props) {
        await onValueChanged(MapService.getValue(data!, prop.dbName)[key],
            prop.overwritePropertiesOnSelect![key]);
      }
    }

    if (await onValueChanged(value, propName)) {
      setState();
    }
  }

  Future<bool> onValueChanged(value, propName, [bool calcAll = false]) async {
    // bool updated = MapService.updateValue(data!, propName, value);
    if (readOnly) return false;
    if (kDebugMode) print('-----------Calc ($propName) $calcAll');
    try {
      if (entity.updatedProps[propName] != null) return false;
      entity.updatedProps[propName] = true;
      bool updated = false;
      if (calcAll) {
        // called onetime on startup
        if (await entity.setDefaultValues(data!)) updated = true;
      } else {
        updated = MapService.updateValue(data!, propName, value);
      }

      // for (Property p in properties) {
      //   var r = await p.isVisible(data, entity);
      //   if (r != p.isVisable) {
      //     p.isVisable = r;
      // updated = true;
      //   }
      // }
      // check the calced properties
      for (PropertyNumber p in calcProps.values
          .where((e) => e.propertyType == PropertyTypes.Number)
          .toList()
          .cast<PropertyNumber>()) {
        if (p.formulaType == "Formula") {
          if (calcAll ||
              p.calcedProps.contains(p.upperFirst(propName)) ||
              p.calcedProps.contains(propName)) {
            var vars = p.calcedProps
                .map((pro) => data![p.lowerFirst(pro)] ?? 0) //TODO: fix dots
                .toList();
            if (await onValueChanged(p.getCalcResult(vars), p.dbName))
              updated = true;
            //setState();
            if (kDebugMode) print('calc ${p.dbName}');
          }
        } else if (p.formulaType == "SumSub") {
          if (calcAll || p.sumEntityName == propName) {
            double val = 0;
            if (data![p.sumEntityName] != null && //TODO: fix dots
                data![p.sumEntityName] is List) {
              //TODO: fix dots
              for (var row in data![p.sumEntityName]) {
                //TODO: fix dots
                val += row[p.sumPropertyName] ?? 0;
              }
            }
            if (await onValueChanged(val, p.dbName)) updated = true;
            //setState();
          }
        }
      }

      // if (updated // || (!calcAll && properties.any((p) => p.filterValueField == propName))
      //     ) setState();
      return updated;
    } catch (err, t) {
      print('XXXXXX ERROR onValueChanged: $err : $t');
    }
    return false;
  }

  Future<bool> onSave() async {
    if (readOnly) return false;
    // show loading
    setLoading();
    // if (!isAddNew && dataOld != null) {
    //   // remove unchanged
    //   for (var k in dataOld!.keys) {
    //     if (k != '_id' && dataOld![k] == data![k]) {
    //       //TODO: fix dots
    //       data!.remove(k);
    //     }
    //   }
    // }
    print(data);
    var result = await EntityService.saveEntity(entity, data, context, isAddNew,
        entitySaveConfirm: _entitySaveConfirm);
    setResume();
    if (!result.success) {
      try {
        if (kDebugMode) print(result.data);
        var question = '';
        if (result.data != null && result.data.length > 0) {
          question = result.data[0]['nameAr'] ?? '';
        }
        if (question.isNotEmpty) {
          var r = await AppNavigator.confirmMessage(
              context, result.message ?? '',
              message: question);
          if (!r) return false;
          result.data[0]['value'] = true;
          _entitySaveConfirm = result.data;
          return await onSave();
        } else {
          AppNavigator.showMessage(context,
              result.message ?? '${result.innerError}', MessageType.error,
              message: question);
          print(result);
        }
      } catch (err, t) {
        await AppNavigator.showMessage(context,
            result.message ?? '${result.innerError}', MessageType.error,
            message: result.message);
        print(err);
        print(t);
      }
      return false;
    }
    ConfigService.setValueMap("savedDataForm${entity.name}", {});
    // show message
    await AppNavigator.showMessage(
        context, 'تم الحفظ بنجاح', MessageType.success);
    AppNavigator.pop(context, true);
    return true;
  }

  Future<bool> onSaveNew() async {
    if (readOnly) return false;
    // show loading
    isAddNew = true;
    resetIds(data!);
    return await onSave();
  }

  Future onOpenHistory() {
    if (data == null) return Future.delayed(Duration.zero);
    return AppNavigator.navigateTo(context,
        () => EntityHistoryPage(entity.dbName, data?['_id'] ?? '', data: data));
  }

  resetIds(Map<String, dynamic> dat) {
    dat['_id'] = null;
    dat['recordNumber'] = null;
    for (var k in dat.keys) {
      if (dat[k] is List) {
        if (kDebugMode) print('d---------$k');
        for (var d in dat[k]) {
          resetIds(d);
        }
      }
    }
  }

  lowerFirst(String s) {
    return '${s[0].toLowerCase()}${s.substring(1)}';
  }

//   void initTabs() {
//     this.tabs = {};
//     for (var prop in this.entity.properties) {
//       if (isAddNew ? prop.isAdd : prop.isEdit) {
//         if (prop.editTabName == null || (prop.editTabName?.isEmpty ?? true))
//           prop.editTabName = AppLocale.translate('Details');

// //prop.propertyType == PropertyTypes.OtherEntityMany
//         if (prop.size == PropertySizes.NewTab) {
//           this.tabs[prop.name] = [
//             [prop]
//           ];
//         } else {
//           if (this.tabs.length == 0)
//             this.tabs[prop.editTabName ?? ''] = [
//               [prop]
//             ];
//           else {
//             if (this.tabs[prop.editTabName] == null) {
//               this.tabs[prop.editTabName ?? ''] = [
//                 [prop]
//               ];
//             } else {
//               if (prop.fullWidth)
//                 this.tabs[prop.editTabName]?.add([prop]);
//               else
//                 this.tabs[prop.editTabName]?.last.add(prop);
//             }
//           }
//         }
//       }
//       if (prop.newLineAfter) this.tabs[prop.editTabName]?.add([]);
//     }

//     for (var tk in tabs.keys) {
//       tabs[tk] = tabs[tk]?.where((e) => e.length > 0).toList() ?? [];
//       if (tabs[tk]?.length == 0) {
//         tabs.remove(tk);
//       }
//     }
//   }
}
